
from .swoopyui import app

from .view import View

from .views.text import Text
from .views.button import Button
from .views.navigationview import NavigationView
from .views.navigationstack import NavigationStack
from .views.navigationlink import NavigationLink
from .views.textfield import TextField


import zipfile


def unzip_file(zip_path, destination_path):
    try:
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(destination_path)
    except FileNotFoundError:
        print("The specified file does not exist.")
    except zipfile.BadZipFile:
        print("The specified file is not a valid ZIP archive.")







class onClientRequestUpdate:
    def __init__(self, update_number, action_name, action_content) -> None:
        self.update_number = update_number
        self.action_name = action_name
        self.action_content = action_content
    
    def load_as_dict (self):
        return {
            "update_number" : self.update_number,
            "action" : {
                "name" : str(self.action_name),
                "content" : self.action_content
            }
        }


class ClientSideUpdate:
    def __init__(self, dict) -> None:
        self.update_name = dict["update_name"]
        self.content = dict["content"]


from ..tools.on_action import on_view_action


class Text (object):
    def __init__(self, text, foreground_color="black", background_color="white", on_hover=None) -> None:
        self.__last_view_id = None # This is used becuase swiftUI will not know that this updated without it
        self.__id = None
        self.__mother_view = None
        self.__parent_view = None

        self.__text = text
        self.__foreground_color = foreground_color
        self.__background_color = background_color
        self.on_hover = on_hover
    

    def get_dict_content (self):
        return {
            "last_view_id" : self.__last_view_id,
            "view_id" : self.__id,
            "vname" : "Text",
            "text" : self.__text,
            "fgcolor" : self.__foreground_color,
            "bgcolor" : self.__background_color
        }

    def respown (self, new_id=None, mother_view=None, parent=None):
        if new_id == None: return
        if mother_view == None: return
        if parent == None: return

        if self.__id == None:
            self.__id = new_id
            self.__last_view_id = new_id
        
        if self.__mother_view == None:
            self.__mother_view = mother_view
        
        if self.__parent_view == None:
            self.__parent_view = parent
    
    def view_action (self, action_data):
        action_name = action_data['action_name']
        if action_name == "on_hover":
            hover_state = action_data['hover_state']
            on_view_action(self.on_hover, [self, {"state":hover_state}])
    
    @property
    def id (self):
        return self.__id

    @property
    def text (self): return self.__text

    @text.setter
    def text (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        self.__text = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id
    
    @property
    def foreground_color (self): return self.__foreground_color

    @foreground_color.setter
    def foreground_color (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        self.__foreground_color = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id


from ..tools.on_action import on_view_action




class NavigationLink (object):
    """A view that controls a navigation presentation."""
    def __init__(self, title:str="Navigate", on_navigate=None) -> None:
        self.__last_view_id = None # This is used becuase swiftUI will not know that this updated without it
        self.__id = None
        self.__mother_view = None
        self.__parent_view = None

        self.on_navigate = on_navigate
        self.__title = title
        self.__subviews = []
    
    def get_dict_content (self):
        all_sub_views = []
        for sv in self.__subviews:
            all_sub_views.append(sv.get_dict_content())
        return {
            "last_view_id" : self.__last_view_id,
            "view_id" : self.__id,
            "vname" : "NavigationLink",
            "title" : self.__title,
            "sub_views" : all_sub_views
        }
    
    def respown (self, new_id=None, mother_view=None, parent=None):
        if new_id == None: return
        if mother_view == None: return
        if parent == None: return

        if self.__id == None:
            self.__id = new_id
            self.__last_view_id = new_id
        
        if self.__mother_view == None:
            self.__mother_view = mother_view
        
        if self.__parent_view == None:
            self.__parent_view = parent
    
    def add (self, *sub_view):
        if self.__mother_view == None:
            raise Exception("Cannot add sub-views while this view not have an active mother view.")
        
        for subv in sub_view:
            subv.respown (new_id=self.__mother_view.get_new_view_id(), mother_view=self.__mother_view, parent=self)
            self.__mother_view.sub_views_history.append(subv)
            self.__subviews.append(subv)
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id
    
    def view_action (self, action_data):
        action_name = action_data['action_name']
        if action_name == "on_navigate":
            on_view_action(self.on_navigate, [self])
    
    def update (self):
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id
        
        self.__parent_view.update()

    @property
    def id (self):
        return self.__id
    

    @property
    def title (self):
        return self.__title
    
    @title.setter
    def title (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        self.__title = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id


from ..tools.on_action import on_view_action




class TextField (object):
    def __init__(self, text:str="", placeholder:str="", width=200, height=40, 
        foreground_color="primery", on_change=None, on_submit=None) -> None:
        self.__last_view_id = None # This is used becuase swiftUI will not know that this updated without it
        self.__id = None
        self.__mother_view = None
        self.__parent_view = None

        self.__text = text
        self.__placeholder = placeholder
        self.__width = width
        self.__height = height
        self.__foreground_color = foreground_color
        self.on_change = on_change
        self.on_submit = on_submit

    def get_dict_content (self):
        return {
            "last_view_id" : self.__last_view_id,
            "view_id" : self.__id,
            "vname" : "TextField",
            "Text" : self.text,
            "placeholder" : self.placeholder,
            "width" : self.width,
            "height" : self.height,
            "fgcolor" : self.foreground_color
        }
    
    def respown (self, new_id=None, mother_view=None, parent=None):
        if new_id == None: return
        if mother_view == None: return
        if parent == None: return

        if self.__id == None:
            self.__id = new_id
            self.__last_view_id = new_id
        
        if self.__mother_view == None:
            self.__mother_view = mother_view
        
        if self.__parent_view == None:
            self.__parent_view = parent


    def view_action (self, action_data):
        action_name = action_data['action_name']
        if action_name == "on_change":
            self.__text = action_data['text']
            on_view_action(self.on_change, [self])
        elif action_name == "on_submit":
            self.__text = action_data['text']
            on_view_action(self.on_submit, [self])

    @property
    def id (self):
        return self.__id
    
    @property
    def text (self):
        return self.__text
    
    @text.setter
    def text (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        self.__text = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id
    

    @property
    def placeholder (self):
        return self.__placeholder
    
    @placeholder.setter
    def placeholder (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        self.__placeholder = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id
    

    @property
    def width (self):
        return self.__width
    
    @width.setter
    def width (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        if not isinstance(value, float) or not isinstance(value, int):
            raise ValueError(f"width must be a number")
        self.__width = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id
    

    @property
    def height (self):
        return self.__height
    
    @height.setter
    def height (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        if not isinstance(value, float) or not isinstance(value, int):
            raise ValueError(f"height must be a number")
        self.__height = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id
    
    @property
    def foreground_color (self): return self.__foreground_color

    @foreground_color.setter
    def foreground_color (self, value):
        if self.__mother_view == None:
            raise Exception("Cannot change the sub_view property while its not on the screen.")
        
        self.__foreground_color = value
        self.__id = self.__mother_view.get_new_view_id()
        self.__mother_view.update(self)
        self.__last_view_id = self.__id








def run_the_target (target_function, function_args, app_class):
    try:
        target_function(*function_args)
    except Exception as e:
        app_class.set_for_the_next_update_get("error", {"error_describe":f"Error: {str(e)}"})
        raise Exception(e)







class View (object):
    def __init__(self, app) -> None:
        self.__app = app
        self.__next_id = 0 #? The next viewID.

        # View props
        self.sub_views_history = [] # This will store all sub_views even from the another subviews.
        self.sub_views = []
    
    def update (self, sub_view=None):
        if sub_view != None:
            action_content = {
                "new_view_content" : sub_view.get_dict_content()
            }
            self.__app.set_for_the_next_update_get (action_name="update_view", action_content=action_content)

    def add (self, view):
        view.respown (new_id=self.get_new_view_id(), mother_view=self, parent=self)
        action_content = {
            "new_view_content" : view.get_dict_content()
        }
        self.__app.set_for_the_next_update_get (action_name="add_view", action_content=action_content)
        self.sub_views_history.append(view)
        self.sub_views.append(view)
    
    def clear (self):
        for sbv in self.sub_views:
            action_content = {
                "view_id" : sbv.id
            }
            self.__app.set_for_the_next_update_get (action_name="delete_view", action_content=action_content)
    
    def delete (self, view):
        """Delete a sub-view from the view"""
        action_content = {
            "view_id" : view.id
        }
        self.__app.set_for_the_next_update_get (action_name="delete_view", action_content=action_content)
        
        num = 0
        for v in self.sub_views_history:
            if v.id == view.id:
                del self.sub_views_history[num]
                break
            num = num + 1

    def get_new_view_id (self):
        """This will generate new viewID for the current and next view"""
        self.__next_id = self.__next_id + 1
        current = self.__next_id
        self.__next_id = self.__next_id + 1
        return current
    
    def manage_on_view_action (self, action_content:dict):
        view_id = action_content['view_id']
        for v in self.sub_views_history:
            if v.id == view_id:
                v.view_action(action_content)
                break


from flask import Flask, request
import socketserver
import tempfile
import threading
import logging
import shutil
import os

from .protocol import onClientRequestUpdate
from .tools.run_target import run_the_target
from .tools.run_swiftUI import run_swiftUI_app
from .view import View


def run_swiftUI_on_new_process (PORT, tmp_dir):
    if not os.path.isdir (tmp_dir):
        print("Cannot found the temp dir.")
        os._exit(0)
    run_swiftUI_app(PORT, tmp_dir)

class app:
    """This is the main function of the app, the function that will run the app and show the window"""
    def __init__(self, target, base_name=__name__) -> None:
        self.__main_view = View(app=self)
        self.__target_function = target
        self.__base_name = base_name
        self.__host()

    def __host (self):
        flask_app = Flask(self.__base_name)

        # Set the logging level to ignore warnings
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

        self.__all_waited_updates = []
        self.__update_number = 0


        @flask_app.route("/", methods=["POST"])
        def index ():
            if self.__all_waited_updates == []:
                return onClientRequestUpdate(
                    self.__update_number, 
                    "", 
                    {"empty":True}).load_as_dict()
            else:
                next_update = self.__all_waited_updates[0]
                action_name = next_update["action"]
                action_content = next_update["action_content"]
                update_number = next_update["update_number"]
                del self.__all_waited_updates[0]
                return onClientRequestUpdate(
                    update_number, 
                    action_name, 
                    action_content).load_as_dict()

        @flask_app.route("/start_the_target_function")
        def start_the_target_function ():
            threading.Thread(target=run_the_target, args=[self.__target_function, [self.__main_view], self], daemon=True).start()
            return ""
        
        @flask_app.route("/client_side_update", methods=["POST"])
        def client_side_update():
            """This will be called on client side update. Like when user click a button this will be called"""
            json_data = request.get_json()
            update_name = json_data.get('update_name')
            if update_name == "on_view_action":
                self.__main_view.manage_on_view_action(json_data.get("update_content"))
            return ""

        
        @flask_app.route("/close_the_app")
        def close_the_app ():
            shutil.rmtree(self.current_tmp_dir)
            os._exit(0)

        with socketserver.TCPServer(("localhost", 0), None) as s:
            free_port = s.server_address[1]
        
        self.current_tmp_dir = tempfile.mkdtemp()
        threading.Thread(target=run_swiftUI_on_new_process, args=[free_port, self.current_tmp_dir]).start()
        flask_app.run(port=free_port)
    
    def set_for_the_next_update_get (self, action_name:str, action_content:dict):
        self.__update_number = self.__update_number + 1
        cont = {"action":f"{action_name}", "action_content":action_content, "update_number":self.__update_number}
        self.__all_waited_updates.append(cont)


from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as f:
    long_des = str(f.read())

setup(
    name='swoopyui',
    version='1.3',
    author='SKbarbon',
    description='A python library that allow you to build swiftUI apps using python.',
    long_description=long_des,
    long_description_content_type='text/markdown',
    url='https://github.com/SKbarbon/swoopyui',
    install_requires=["flask", "requests"],
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ],
    include_dirs=["assets"],
    package_data={"assets": ["swoopyui.zip"]},
    include_package_data=True
)



import subprocess
import os
import shutil
from .unzip_assets import unzip_file

def run_swiftUI_app(port, tmp_dir):
    temp_dir = tmp_dir
    print(f"tmp exist: {os.path.isdir(temp_dir)}")

    zip_path = str(__file__).replace("tools/run_swiftUI.py", "assets/swoopyui.zip")
    new_app_path = os.path.join(temp_dir, "swoopyui.app/")

    unzip_file(zip_path=zip_path, destination_path=temp_dir)

    executable_of_the_app = os.path.join(new_app_path, "Contents/MacOS/swoopyui")
    chmod_command = ["chmod", "+x", executable_of_the_app]

    subprocess.run(chmod_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    run_command = [executable_of_the_app, str(port)]
    subprocess.run(run_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    print(temp_dir)
    shutil.rmtree(temp_dir)